#######################################################################
# Copyright (c) 2017 Josef Spidlen, Ph.D., FlowJo, LLC
#
# License
# The software is distributed under the terms of the 
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
# 
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied 
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct, 
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or 
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this 
# software.
######################################################################

suppressMessages(library("FlowSOM"))
suppressMessages(library("flowCore"))

sessionInfo()
# We want FlowSOM >= 1.8.0 (seed only works for FlowSOM >= 1.8.0)

## Just for debugging to see what FlowJo is actually giving us.
##
# cat("FJ-DATA-FILE-PATH   = FJ_DATA_FILE_PATH \n")
# cat("FJ-CSV-OUPUT-FILE   = FJ_CSV_OUPUT_FILE \n")
# cat("FJ-PAR-NUM-CLUST    = FJ_PAR_NUM_CLUST \n")
# cat("FJ-PARAMS-LIST      = FJ_PARAMS_LIST \n")
# cat("FJ-BG-META-CLUSTERS = FJ_BG_META_CLUSTERS \n")
# cat("FJ-PLOT-OPTION      = FJ_PLOT_OPTION \n")
# cat("FJ-VIEW-OPTION      = FJ_VIEW_OPTION \n")
# cat("FJ-PLOT-ALL         = FJ_PLOT_ALL \n")

plotAllAsStars <- FJ_PLOT_ALL
bgMetaClusters <- FJ_BG_META_CLUSTERS
plotView <- "FJ_VIEW_OPTION"
plotPar <- "FJ_PLOT_OPTION"

dataset <- 1

robust.read.FCS <- function(fcsPath) {
	out <- tryCatch(
			{
				read.FCS(fcsPath, dataset=dataset)
			}, 
			error=function(cond) {
				read.FCS(fcsPath, dataset=dataset, emptyValue=FALSE, ignore.text.offset=TRUE)
			})    
	return(out)
}

inputFCS <- robust.read.FCS("FJ_DATA_FILE_PATH");

eventsCount <- length(inputFCS@exprs[,1])

parNames <- c(FJ_PARAMS_LIST)

if (length(parNames) == 0)
	stop("Some input parameters need to be selected!", call.=FALSE)

# FlowJo sometimes gives us names like [V705-A] but the actual names in the FCS file are <V705-A>
# FlowJo also replaces '_' with '/' so let's try fix that as well
parNames <- unlist(lapply(parNames, function(name) {
					if (name %in% colnames(inputFCS)) {
						name
					} else {
						# Let's try [] to <>
						name2 <- gsub("[", "<", name, fixed=TRUE)
						name2 <- gsub("]", ">", name2, fixed=TRUE)
						if (name2 %in% colnames(inputFCS)) {
							name2 # Worked, return it
						} else {
							# Previous fix did not do it, _ => / on the original names
							name2 <- gsub("_", "/", name, fixed=TRUE)
							if (name2 %in% colnames(inputFCS)) {
								name2 # Worked, return it
							} else {
								# That did not work either, let's try both [] => on top of the previous fix (_ => /)
								name3 <- gsub("[", "<", name2, fixed=TRUE)
								name3 <- gsub("]", ">", name3, fixed=TRUE)
								if (name3 %in% colnames(inputFCS)) {
									name3 # Worked, finally, return it
								} else {
									## Maybe we read a wrong dataset?
									cat(paste("The input FCS file does not contain the provided input parameter, missing", name, "\n"))
									"TRYDIFFERENTDATASET"
								}
							}
						}
					}
				}))

## DEBUG info

inputFCS

parNames

eventsCount

if (length(parNames) == 0 || "TRYDIFFERENTDATASET" %in% colnames(inputFCS) || !("time" %in% tolower(colnames(inputFCS))))
{
	## Maybe we read the wrong dataset? Let's try the second one
	inputFCS <- new("flowFrame")
	dataset <- 2
	inputFCS <- robust.read.FCS("FJ_DATA_FILE_PATH");
	eventsCount <- length(inputFCS@exprs[,1])
	parNames <- c(FJ_PARAMS_LIST)
	
	parNames <- unlist(lapply(parNames, function(name) {
						if (name %in% colnames(inputFCS)) {
							name
						} else {
							# Let's try [] to <>
							name2 <- gsub("[", "<", name, fixed=TRUE)
							name2 <- gsub("]", ">", name2, fixed=TRUE)
							if (name2 %in% colnames(inputFCS)) {
								name2 # Worked, return it
							} else {
								# Previous fix did not do it, _ => / on the original names
								name2 <- gsub("_", "/", name, fixed=TRUE)
								if (name2 %in% colnames(inputFCS)) {
									name2 # Worked, return it
								} else {
									# That did not work either, let's try both [] => on top of the previous fix (_ => /)
									name3 <- gsub("[", "<", name2, fixed=TRUE)
									name3 <- gsub("]", ">", name3, fixed=TRUE)
									if (name3 %in% colnames(inputFCS)) {
										name3 # Worked, finally, return it
									} else {
										stop(paste("The input FCS file does not contain the provided input parameters, missing", name), call.=FALSE)
									}
								}
							}
						}
					}))
}

## DEBUG info

inputFCS

parNames

eventsCount

dataset

if (!plotAllAsStars) {
	## If not plotting all then make sure plotPar is among parameters in the FCS file
	if (!(plotPar %in% colnames(inputFCS))) {
		plotPar2 <- gsub("[", "<", plotPar, fixed=TRUE)
		plotPar2 <- gsub("]", ">", plotPar2, fixed=TRUE)
		if (plotPar2 %in% colnames(inputFCS)) {
			plotPar <- plotPar2
		} else {
			plotPar2 <- gsub("_", "/", plotPar, fixed=TRUE)
			if (plotPar2 %in% colnames(inputFCS)) {
				plotPar <- plotPar2
			} else {
				plotPar3 <- gsub("[", "<", plotPar2, fixed=TRUE)
				plotPar3 <- gsub("]", ">", plotPar3, fixed=TRUE)
				if (plotPar3 %in% colnames(inputFCS)) {
					plotPar <- plotPar3
				} else {
					stop(paste("The input FCS file does not contain ", plotPar), call.=FALSE)
				}
			} 
		}
	}
}

if (eventsCount == 0)
	stop("R failed to read the input FCS file.", call.=FALSE)

parIndices <- match(parNames,colnames(inputFCS))

if (length(parNames) == 0 || length(parIndices) == 0)
	stop("Something seems wrong, it's like the input FCS file does not contain the provided input parameters.", call.=FALSE)

fSOM <- FlowSOM(
  inputFCS,
  compensate = FALSE,
  transform = TRUE, 
  toTransform=parIndices,
  scale = TRUE,
  colsToUse = parIndices, 
  xdim = length(parIndices), 
  ydim = length(parIndices),
  nClus = FJ_PAR_NUM_CLUST,
  seed = 12345)

labels <- as.matrix(as.numeric((fSOM$metaclustering[fSOM$FlowSOM$map$mapping[,1]])) * 100 + runif(min=-1, max=1, n=eventsCount));
colnames(labels) <- c("FlowSOM")
write.csv(labels, file="FJ_CSV_OUPUT_FILE", row.names=FALSE)

png(file=paste0("FJ_CSV_OUPUT_FILE", ".png"), bg="transparent", width = 1280, height = 1280, units = "px", pointsize = 24)

if (plotAllAsStars) {
	## Use PlotStars
	if (bgMetaClusters) {
		PlotStars(
			fSOM$FlowSOM,
			view=plotView,
			backgroundValues = as.factor(as.numeric(fSOM$metaclustering) - 1)) ## -1 to make it go 0..n-1 instead 1..n to match FlowJo's populations
	} else {
		PlotStars(
			fSOM$FlowSOM,
			view=plotView)
	}
} else {
	## Use PlotMarker
	if (bgMetaClusters) {
		PlotMarker(
			fSOM$FlowSOM,
			plotPar,
			view=plotView,
			backgroundValues = as.factor(as.numeric(fSOM$metaclustering) - 1)) ## -1 to make it go 0..n-1 instead 1..n to match FlowJo's populations
	} else {
		PlotMarker(
			fSOM$FlowSOM,
			plotPar,
			view=plotView)
	}	
}


dev.off()


